import json
from json import JSONDecodeError
from typing import List, Dict

import click
from click import UNPROCESSED
from jsonschema import validate, ValidationError

from src import LOGGER, PHOENIX_SCHEMA
from src.core.exceptions import (
    InvalidTemplateException,
    CommandNotFoundException,
    ActionNotFoundException,
    ShowHelpException,
    PhoenixException,
    PhoenixWarningException,
)
from src.core.models import Execution
from src.core.px_git import require_git_repo
from src.core.px_questionary import select
from src.core.rules import fire_rules
from src.core.utils import get_template, px_print


@click.command(
    context_settings=dict(
        allow_extra_args=True,
        ignore_unknown_options=True,
    )
)
@click.version_option(
    package_name="git-phoenix",
    prog_name="Git Phoenix",
)
@click.argument("args", nargs=-1, type=UNPROCESSED)
def main(args):
    try:
        px_print("\n-#- P H O E N I X  B E G I N -#-\n")

        LOGGER.info("Iniciando processamento...")

        require_git_repo()
        template = _get_template()

        try:
            validate(instance=template, schema=PHOENIX_SCHEMA)

            LOGGER.info("Validando comando...")
            command = _get_command(args=args, template=template)
            LOGGER.info("Validando ação...")
            action = _get_action(args=args, command=command)
            variables = template.get("variables")
            arguments = args[2:]

            # Create execution object to carry template and
            # arguments through execution
            execution = Execution(
                command=command,
                action=action,
                variables=variables,
                arguments=arguments,
            )

            LOGGER.info("Executando regras...")

            fire_rules(execution)
        except ValidationError as e:
            raise InvalidTemplateException(e.message)
    except PhoenixWarningException as e:
        LOGGER.warn(e.message)
    except PhoenixException as e:
        LOGGER.exception(e.message)
    except Exception as e:
        LOGGER.exception(str(e))

    px_print("\n-#- P H O E N I X  E N D -#-")


def _is_argument(arg):
    return arg.startswith("-")


def _get_template():
    template_path = get_template()

    with open(template_path, encoding="utf-8") as template_file:
        try:
            return json.load(template_file)
        except JSONDecodeError as e:
            raise InvalidTemplateException(
                f"{e.msg} - Error at line:{e.lineno} and column: {e.colno}"
            )


def _validate_if_init(args, template):
    if "init" == args[0] and {} != template.get("init"):
        raise InvalidTemplateException()


def _get_commands(template) -> List[str]:
    choices = []

    if {} != template.get("init"):
        choices.append("init")

    choices.extend(command.get("name") for command in template.get("commands"))

    return choices


def _get_command_by_template(template) -> Dict:
    command = select(
        msg="Escolha um dos comandos abaixo:", choices=_get_commands(template)
    )

    return _get_command_by_name(command, template)


def _get_command_by_name(command, template):
    return next(
        filter(
            lambda x: x.get("name") == command,
            template.get("commands"),
        ),
        None,
    )


def _get_actions(command) -> Dict:
    return {action.get("name"): action for action in command.get("actions")}


# def _get_actions_alias(command) -> List[str]:
#     return [action.get("alias") for action in command.get("actions")]


def _get_action_by_command(command) -> Dict:
    actions = _get_actions(command)
    action = select(
        msg="Escolha uma das ações abaixo:", choices=actions.keys()
    )

    return actions.get(action)


def _get_action_by_name(action, command):
    return next(
        filter(
            lambda x: x.get("name") == action,
            command.get("actions"),
        ),
        None,
    )


def _passed_command(args: List[str]) -> bool:
    return len(args) > 0


def _get_command_by_argument(command: str, template: Dict):
    commands = _get_commands(template)

    if command not in commands:
        raise CommandNotFoundException(command)
    else:
        return _get_command_by_name(command, template)


def _passed_action(args: List[str]) -> bool:
    return len(args) > 1


def _get_action_by_argument(action: str, command: Dict):
    actions = _get_actions(command)
    # actions_alias = _get_actions_alias(command)

    if action not in actions:
        raise ActionNotFoundException(action)
    else:
        return _get_action_by_name(action, command)


def _get_command(args, template):
    if _passed_command(args):
        command = _get_command_by_argument(args[0], template)

        if len(args) > 1 and "-h" == args[1]:
            raise ShowHelpException(command.get("help"))

        return command
    else:
        return _get_command_by_template(template)


def _get_action(args, command):
    if _passed_action(args):
        action = _get_action_by_argument(args[1], command)

        if len(args) > 2 and "-h" == args[2]:
            raise ShowHelpException(action.get("help"))

        return action
    else:
        return _get_action_by_command(command)



import git
from git import GitCommandError, Repo

from src import LOGGER
from src.core.exceptions import (
    NotAGitRepoException,
    BranchAlreadyExistsException,
    GitException,
)

GIT_REPO: Repo


def get_config(key):
    try:
        global GIT_REPO
        return GIT_REPO.git.config(key)
    except GitCommandError:
        return f"Config {key} not found"


def get_tags():
    global GIT_REPO
    return [
        version.replace("v", "") for version in GIT_REPO.git.tag().split("\n")
    ]


def require_git_repo():
    if not _is_git_repo():
        raise NotAGitRepoException()


def _is_git_repo():
    try:
        global GIT_REPO
        GIT_REPO = Repo(".", search_parent_directories=True)
        return True
    except git.exc.InvalidGitRepositoryError:
        return False


def has_unstaged_files():
    unstaged_files = GIT_REPO.git.status("--porcelain")

    if not unstaged_files:
        return False

    return True


def is_ahead():
    branch = retrieve_current_branch()

    try:
        commits_ahead = GIT_REPO.iter_commits(branch + ".." + branch)
        number_of_commits = sum(1 for _ in commits_ahead)

        return number_of_commits > 0
    except GitCommandError:
        LOGGER.info(
            f"Your current source ({branch}) doesn't exists on remote "
            f"repo. Please use git push source {branch}."
        )


def retrieve_current_branch():
    return GIT_REPO.active_branch.name


def checkout(branch):
    global GIT_REPO

    LOGGER.info(f"Fazendo checkout da source {branch}...")

    GIT_REPO.git.checkout(branch)


def _validate_existence(branch):
    global GIT_REPO
    remotes = [branch.name.replace("origin/", "") for branch in GIT_REPO.refs]
    local = GIT_REPO.branches

    if branch in remotes or branch in local:
        raise BranchAlreadyExistsException()


def checkout_new_branch(source: str, branch: str):
    global GIT_REPO

    checkout(source)
    pull(source)

    LOGGER.info(f"Criando source {branch} com base na source {source}...")

    try:
        _validate_existence(branch)

        GIT_REPO.git.checkout(source, b=branch)
    except GitCommandError as e:
        raise GitException(e.stdout)


def merge(source: str, target: str, allow_merge_again: bool):
    global GIT_REPO

    if allow_merge_again or not _already_merged(target, source):
        checkout(target)
        pull(target)

        LOGGER.info(
            f"Realizando merge da source {source} com a source {target}..."
        )
        try:
            GIT_REPO.git.merge(source, "--no-ff")
        except GitCommandError:
            raise GitException(
                "O merge gerou conflitos, você precisa "
                "resolvê-los antes de commitar!"
            )
    else:
        LOGGER.warn(
            f"O merge da source {source} com a source {target} já "
            f"foi realizado!"
        )


def merge_request(source: str, target: str, allow_merge_again: bool,
                  mr_template: str):
    pass  # this method will be implemented on version 1.1.0


def pull(branch):
    global GIT_REPO

    LOGGER.info(f"Atualizando source {branch}...")

    try:
        GIT_REPO.git.pull()
    except GitCommandError as e:
        raise GitException(e.stdout)


def delete(pattern):
    LOGGER.info(f"Excluindo source(es) {pattern}...")

    branches = [
        branch.replace(" ", "") for branch in git.branch().splitlines()
    ]

    branches = list(filter(lambda x: x.startswith(pattern), branches))

    [git.execute(["git", "source", "-D", branch]) for branch in branches]


def _already_merged(destination, branch):
    merged_branches = log(
        ["--oneline", "--merges", "--grep=into", destination]
    ).splitlines()

    for merged_branch in merged_branches:
        translated_merge_message = _translate_merge_message(msg=merged_branch)

        if translated_merge_message["source"] == branch:
            return True

    return False


def log(parameters):
    git = Repo(".", search_parent_directories=True).git

    if parameters:
        return git.log(parameters)

    return git.log()


def _translate_merge_message(msg):
    translated_merge_message = {}

    origin = msg.split("source ")

    if origin[1:]:
        origin = origin[1]
    else:
        origin = origin[0]

    origin = origin[: origin.index(" ")]

    translated_merge_message["source"] = origin.replace("'", "")
    translated_merge_message["target"] = msg.split("into ")[-1].replace(
        "'", ""
    )

    return translated_merge_message



import importlib.resources as lib_resources
import json
import logging

import coloredlogs
from questionary import Style

from . import resources

# initialize logger
LOGGER = logging.getLogger("Git Phoenix")

# parse phoenix-schema
PHOENIX_SCHEMA = json.loads(
    lib_resources.read_text(resources, "phoenix-schema.json")
)

# refs.: https://stackoverflow.com/a/20112491/7973282
fmt = (
    "%(asctime)s :: %(name)s :: %(levelname)s - "
    "%(message)s (%(filename)s:%(lineno)d - %(funcName)s())"
)

coloredlogs.DEFAULT_FIELD_STYLES = {
    "asctime": {"color": "white"},
    "levelname": {"color": "black", "bold": True},
    "name": {"color": "red", "bold": True},
}

coloredlogs.DEFAULT_LEVEL_STYLES = {
    "info": {"color": "cyan"},
    "notice": {"color": "magenta"},
    "verbose": {"color": "blue"},
    "success": {"color": "green", "bold": True},
    "spam": {"color": "blue"},
    "critical": {"color": "red", "bold": True},
    "error": {"color": "red"},
    "debug": {"color": "green"},
    "warning": {"color": "yellow"},
}

coloredlogs.install(level="INFO", logger=LOGGER, fmt=fmt, milliseconds=True)

CUSTOM_STYLE = Style(
    [
        ("qmark", "ansibrightyellow bold"),  # token in front of the question
        ("question", "ansiwhite bold"),  # question text
        (
            "answer",
            "ansibrightmagenta bold",
        ),  # submitted answer text behind the question
        (
            "pointer",
            "ansibrightblue bold",
        ),  # pointer used in select and checkbox prompts
        (
            "highlighted",
            "ansibrightblack bold",
        ),  # pointed-at choice in select and checkbox prompts
        (
            "selected",
            "ansibrightblack",
        ),  # style for a selected item of a checkbox
        ("separator", "ansibrightmagenta"),  # separator in lists
        (
            "instruction",
            "",
        ),  # user instructions for select, rawselect, checkbox
        ("text", ""),  # plain text
        (
            "disabled",
            "ansibrightblack italic",
        ),  # disabled choices for select and checkbox prompts
    ]
)



from dataclasses import dataclass

from src.core.actions.executable import Executable
from src.core.actions.executable import _validate_pattern
from src.core.models import ActionExecution
from src.core.px_git import checkout_new_branch
from src.core.px_questionary import confirm
from src.core.template_models import Branch, Affix, Pattern


@dataclass
class CreateBranchParameters:
    name: str
    source: Branch
    affix: Affix
    pattern: Pattern


class CreateBranch(Executable):
    parameters: CreateBranchParameters

    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)
        action_parameters = self.action_execution.parameters

        self.parameters = CreateBranchParameters(
            name=action_parameters.get("name", ""),
            source=Branch(action_parameters.get("source", {})),
            affix=Affix(action_parameters.get("affix", {})),
            pattern=Pattern(action_parameters.get("pattern", {})),
        )

    def execute(self):
        name = self.parameters.name
        source = self.parameters.source
        affix = self.parameters.affix
        pattern = self.parameters.pattern

        _validate_pattern(source.pattern, source.name, "Source name invalid")
        _validate_pattern(pattern, name, "Name invalid")

        if affix:
            final_name = [name]

            if affix.prefix:
                final_name = affix.prefix + final_name
                name = affix.join_char.join(final_name)
            if affix.suffix:
                final_name.extend(affix.suffix)
                name = affix.join_char.join(final_name)

        confirmed = confirm(
            msg=f"Você confirma a criação da "
            f"branch {name} com base na "
            f"branch {source.name}?"
        )

        if confirmed:
            checkout_new_branch(source=source.name, branch=name)



from typing import List, TextIO

from prompt_toolkit.output.win32 import NoConsoleScreenBufferError, Win32Output
from questionary import form as f, select as s, confirm as c, text as t

from core.utils import px_print
from src import CUSTOM_STYLE
from src.core.utils import raiser
from src.core.exceptions import (
    ProcessCancelledException,
    InvalidOptionException,
)

IS_VALID_ENV = True

try:
    Win32Output(stdout=TextIO()).get_win32_screen_buffer_info()
except NoConsoleScreenBufferError:
    IS_VALID_ENV = False


def select(msg: str, choices: List[str]):
    if IS_VALID_ENV:
        answer = f(
            answer=s(
                msg,
                choices=choices,
                style=CUSTOM_STYLE,
            )
        ).ask(kbi_msg="")

        return _return(answer)
    else:
        px_print(f"? {msg}")

        for choice in choices:
            px_print(f"» {choice}")

        answer = _input()

        if answer in choices:
            return answer
        else:
            raise InvalidOptionException(answer, choices)


def confirm(msg: str):
    answer = c(message=msg, style=CUSTOM_STYLE).ask(kbi_msg="")
    return _return(answer)


def text(msg: str):
    answer = t(message=msg, qmark="●", style=CUSTOM_STYLE).ask(kbi_msg="")
    return _return(answer)


def _return(answer):
    return (
        answer.get("answer")
        if answer
        else raiser(
            ProcessCancelledException("Usuário cancelou o processamento!")
        )
    )


def _input():
    try:
        return input("● Qual opção você deseja?")
    except (KeyboardInterrupt, EOFError):
        return raiser(
            ProcessCancelledException("Usuário cancelou o processamento!")
        )



from unittest import TestCase

from src.core.utils import capitalize_first_letter


class TestUtils(TestCase):
    def test_capitalize_first_letter(self):
        text = "Testing capitalize"
        text_to_test = capitalize_first_letter("testing capitalize")

        assert text == text_to_test



from abc import ABC
from dataclasses import dataclass
from typing import List, Type, Dict

from src.core.enums import Action


@dataclass
class Affix:
    prefix: List[str]
    suffix: List[str]
    join_char: str

    def __init__(self, affix: Dict):
        self.prefix = affix.get("prefix", [])
        self.suffix = affix.get("suffix", [])
        self.join_char = affix.get("join_char", "/")


@dataclass
class Pattern:
    regex: str
    message: str
    example: str

    def __init__(self, pattern: Dict):
        self.regex = pattern.get("regex", "")
        self.message = pattern.get("message", "")
        self.example = pattern.get("example", "")


@dataclass
class Branch:
    name: str
    pattern: Pattern

    def __init__(self, branch: Dict):
        self.name = branch.get("name", "")
        self.pattern = Pattern(branch.get("pattern", {}))


@dataclass
class Parameters(ABC):
    pass


@dataclass
class Do:
    action: Type[Action]
    parameters: Type[Parameters]


@dataclass
class Execution:
    step: int
    do: Type[Do]


@dataclass
class Action:
    name: str
    alias: str
    execution: List[Execution]


@dataclass
class Command:
    name: str
    alias: str
    actions: List[Action]


@dataclass
class Init:
    execution: List[Execution]


@dataclass
class Template:
    init: Type[Init]
    commons: Dict
    commands: List[Command]



from distutils.version import LooseVersion

from src.core.px_git import get_tags
from src.core.enums import TagIncrement
from src.core.px_questionary import text


def user_input(msg: str, execute: bool = True):
    if execute:
        return text(msg=msg.replace("%", ""))
    else:
        input_variable_name = "o parâmetro"

        if msg.count("%") == 2:
            input_variable_name = msg.split("%")[1]

        return f"Você quer digitar {input_variable_name}?"


def get_latest_version(increase: str, execute: bool = True):
    if execute:
        latest_tag = max(get_tags(), key=LooseVersion)
        version = latest_tag.split(".")
        tag_increment = TagIncrement(increase.lower())

        if tag_increment == TagIncrement.MAJOR:
            version[0] = str(int(version[0]) + 1)
            version[1] = "0"
            version[2] = "0"

        elif tag_increment == TagIncrement.MINOR:
            version[1] = str(int(version[1]) + 1)
            version[2] = "0"

        else:
            version[2] = str(int(version[2]) + 1)

        return ".".join(version)
    else:
        return "Identifique a última tag para mim"


def get_latest_tag(increase: TagIncrement, execute: bool = True):
    pass  # this method will be implemented on version 1.1.0


def get_branch(pattern: str, execute: bool = True):
    pass  # this method will be implemented on version 1.1.0


def get_origin_branch(execute: bool = True):
    pass  # this method will be implemented on version 1.1.0



from dataclasses import dataclass
from typing import List


from src.core import merge, merge_request
from src.core.actions.executable import (
    Executable,
    _validate_pattern,
    _validate_branch_patterns,
)
from src.core.models import ActionExecution
from src.core.px_questionary import confirm
from src.core.template_models import Branch


@dataclass
class MergeParameters:
    source: Branch
    targets: List[Branch]
    allow_new_merge: bool


@dataclass
class MergeRequestParameters:
    source: Branch
    target: Branch
    mr_template: str


class Merge(Executable):
    parameters: MergeParameters

    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)
        action_parameters = self.action_execution.parameters

        self.parameters = MergeParameters(
            source=Branch(action_parameters.get("source", {})),
            targets=[
                Branch(branch)
                for branch in action_parameters.get("targets", {})
            ],
            allow_new_merge=action_parameters.get("allow_new_merge", {}),
        )

    def execute(self):
        source = self.parameters.source
        targets = self.parameters.targets
        allow_new_merge = self.parameters.allow_new_merge

        _validate_pattern(source.pattern, source.name, "Source name invalid")
        _validate_branch_patterns(targets, "Target name invalid")

        confirmed = confirm(
            msg=f"Você confirma o merge da "
            f"branch {source.name} com a(s) branch(es) "
            f"{[','.join(branch.name) for branch in targets]}?"
        )

        if confirmed:
            for target in targets:
                targets.pop(0)
                merge(source.name, target.name, allow_new_merge)


class MergeRequest(Executable):
    parameters: MergeRequestParameters

    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)
        action_parameters = self.action_execution.parameters

        self.parameters = MergeRequestParameters(
            source=Branch(action_parameters.get("source", {})),
            target=Branch(action_parameters.get("target", {})),
            mr_template=action_parameters.get("mr_template", {}),
        )

    def execute(self):
        source = self.parameters.source
        target = self.parameters.target
        mr_template = self.parameters.mr_template

        _validate_pattern(source.pattern, source.name, "Source name invalid")
        _validate_pattern(target.pattern, target.name, "Target name invalid")

        confirmed = confirm(
            f"Você confirma abrir o merge request da "
            f"branch {source.name} com a(s) branch(es) "
            f"{target.name}?"
        )

        if confirmed:
            merge_request(source.name, target.name, True, mr_template)



from src import LOGGER
from src.core.exceptions import UnstagedFilesException, BranchAheadException
from src.core.px_git import has_unstaged_files, is_ahead
from src.core.models import Execution, ActionExecution
from src.core.utils import capitalize_first_letter
from src.core.actions.block_commit import BlockCommit
from src.core.actions.create_branch import CreateBranch
from src.core.actions.custom_actions import CustomActions
from src.core.actions.delete_branch import DeleteBranch
from src.core.actions.merge import Merge
from src.core.actions.merge import MergeRequest
from src.core.actions.rebase import Rebase
from src.core.actions.tag import Tag


def fire_rules(execution: Execution):
    if has_unstaged_files():
        raise UnstagedFilesException()

    if is_ahead():
        raise BranchAheadException()

    actions = []

    for action_execution in execution.action.get("execution"):
        do = action_execution.get("do")
        action_from_template = do.get("action")

        LOGGER.info(f"Validando método {action_from_template}...")

        action_class = eval(capitalize_first_letter(action_from_template))

        action = action_class(
            ActionExecution(
                execution.variables, do.get("parameters"), execution.arguments
            )
        )

        if not action.is_implemented:
            LOGGER.warn("Método não implementado, favor revisar seu template!")
        else:
            actions.append(action)

    # TODO implementar informativo quando a ação não puder ser executada por
    #  determinado motivo

    for action in actions:
        action.execute()



from dataclasses import dataclass
from typing import List

from src import LOGGER
from src.core.actions.executable import Executable
from src.core.enums import TagIncrement, TagReference
from src.core.models import ActionExecution
from src.core.template_models import Branch


@dataclass
class TagParameters:
    reference: TagReference
    increment: TagIncrement
    targets: List[Branch]


class Tag(Executable):
    parameters: TagParameters

    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)
        action_parameters = self.action_execution.parameters

        self.parameters = TagParameters(
            reference=action_parameters.get("reference"),
            increment=action_parameters.get("increment"),
            targets=[
                Branch(branch)
                for branch in action_parameters.get("targets", {})
            ],
        )

    def execute(self):
        LOGGER.warn("Not implemented yet!")



from dataclasses import dataclass
from typing import List, Dict


@dataclass
class Execution:
    command: Dict
    action: Dict
    variables: Dict
    arguments: List[str]


@dataclass
class ActionExecution:
    variables: Dict
    parameters: Dict
    arguments: List[str]


@dataclass
class Choice:
    index: int = 0
    choice_text: str = ""
    text: str = ""



from src import LOGGER
from src.core.actions.executable import Executable
from src.core.models import ActionExecution


class BlockCommit(Executable):
    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)
        self.is_implemented = False

    def execute(self):
        LOGGER.debug("Not implemented yet!")



from enum import Enum, EnumMeta


class MetaEnum(EnumMeta):
    def __contains__(self, item):
        try:
            self(item)
        except ValueError:
            return False
        return True


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


class Action(BaseEnum):
    CREATE_BRANCH = "createBranch"
    DELETE_BRANCH = "deleteBranch"
    MERGE = "merge"
    MERGE_REQUEST = "merge_request"
    TAG = "tag"


class TagReference(BaseEnum):
    BRANCH = "source"
    TAG = "tag"


class TagIncrement(BaseEnum):
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"



from src import LOGGER
from src.core.models import ActionExecution
from src.core.actions.executable import Executable


class Rebase(Executable):

    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)

    def execute(self):
        LOGGER.warn("Not implemented yet!")



from abc import ABC
from typing import List, Dict

from regex import regex

from src import LOGGER
from src.core import template_methods
from src.core.exceptions import (
    InvalidPatternException,
    InvalidVariableException,
    InvalidTemplateException,
    MethodNotImplementedException,
)
from src.core.models import ActionExecution, Choice
from src.core.px_questionary import select
from src.core.template_models import Pattern, Branch
from src.core.utils import list_functions

AVAILABLE_METHODS = list_functions(template_methods)


def _adjust_regex(r: str) -> str:
    if not r.startswith("^"):
        r = "^" + r
    if not r.endswith("$"):
        r += "$"
    return r.replace("/", "\\/")


def _validate_pattern(pattern: Pattern, text: str, msg: str):
    if pattern.regex:
        r = _adjust_regex(pattern.regex)
        p = regex.compile(r)

        if not p.match(text):
            if pattern.message:
                msg = pattern.message

            raise InvalidPatternException(
                f"Invalid text: {text} | {msg} | Example: {pattern.example}"
            )


def _validate_branch_patterns(branches: List[Branch], msg: str):
    for branch in branches:
        _validate_pattern(branch.pattern, branch.name, msg)


def _is_regex(value: str) -> bool:
    return value.startswith("^") and value.endswith("$")


def _ismethod(value: str) -> bool:
    return value.startswith("@")


def _isvariable(value: str) -> bool:
    return "$" in value and not value.endswith("$")


def _ismultiplechoice(value: str) -> bool:
    return " or " in value


class Executable(ABC):
    def __init__(self, action_execution: ActionExecution):
        self.action_execution = action_execution
        self.is_implemented = True
        self.index_method_executed = 0
        self._parse_action_parameters()

    def execute(self):
        raise NotImplementedError("This is an abstract method!")

    def _parse_action_parameters(self):
        LOGGER.info("Realizando parse da ação no template...")

        self.action_execution.parameters = {
            k: self._change_value(v)
            for k, v in self.action_execution.parameters.items()
        }

    def _search_variable(self, variable: str) -> str:
        try:
            return self.action_execution.variables[variable]
        except KeyError:
            raise InvalidVariableException(
                f"Variable {variable} not found on " "template"
            )

    def _change_value(self, value, execute: bool = True):
        if isinstance(value, str):
            return self._process_str(value, execute)
        elif isinstance(value, list):
            return self._process_list(value, execute)
        elif isinstance(value, dict):
            return self._process_dict(value, execute)

        return value

    def _process_str(self, value: str, execute: bool = True) -> str:
        LOGGER.debug(f"Processando parse de string {value}...")

        if _ismultiplechoice(value):
            return self._process_multiple_choice(value)
        elif _ismethod(value):
            return self._process_method(value, execute)
        elif _isvariable(value):
            return self._process_variable(value)

        return value

    def _process_list(self, value: List, execute: bool = True) -> List:
        LOGGER.debug(f"Processando parse de lista {value}...")

        return [self._change_value(item, execute) for item in value]

    def _process_dict(self, value: Dict, execute: bool = True) -> Dict:
        LOGGER.debug(f"Processando parse de dicionário {value}...")

        return {k: self._change_value(v, execute) for k, v in value.items()}

    def _process_multiple_choice(self, value: str) -> str:
        LOGGER.debug(f"Processando parse de escolhas {value}...")

        begin = 0
        end = len(value)

        if value.startswith("("):
            begin = 1
        if value.endswith(")"):
            end = -1

        value = value[begin:end]

        if "user_input" in value and self.action_execution.arguments:
            index = self.index_method_executed
            self.index_method_executed += 1
            return self.action_execution.arguments[index]

        choices = {}
        choice_index = 1

        for choice in value.split(" or "):
            choices[str(choice_index)] = Choice(
                choice_index, choice, self._change_value(choice, execute=False)
            )

            choice_index += 1

        choice = select(
            "Escolha uma das opções abaixo:",
            choices=[f"{k}. {v.text}" for k, v in choices.items()],
        )

        user_choice = choices.get(
            choice.get("answer").split(".")[0]
        ).choice_text

        return self._change_value(user_choice)

    def _process_method(self, value: str, execute: bool = True) -> str:
        LOGGER.debug(f"Processando parse método {value}...")

        method_definition = value.split("@")[1]
        method_name = method_definition[0 : method_definition.find("(")]

        if method_name not in AVAILABLE_METHODS:
            raise MethodNotImplementedException()

        if "user_input" == method_name and self.action_execution.arguments:
            return self.action_execution.arguments[self.index_method_executed]

        original_arguments = method_definition[
            method_definition.find("(") + 1 : method_definition.find(")")
        ]

        # refs.: https://stackoverflow.com/a/48838456/7973282
        parsed_arguments = eval("dict({})".format(original_arguments))

        parsed_arguments = {
            k: self._change_value(v) for k, v in parsed_arguments.items()
        }

        parsed_arguments["execute"] = execute

        parsed_arguments = ",".join(
            [
                f"{k}='{v}'" if isinstance(v, str) else f"{k}={v}"
                for k, v in parsed_arguments.items()
            ]
        )

        method_definition = f"{method_name}({parsed_arguments})"

        try:
            return eval(f"template_methods.{method_definition}")
        except TypeError as e:
            LOGGER.error(str(e))
            raise InvalidTemplateException()

    def _process_variable(self, value: str) -> str:
        LOGGER.debug(f"Processando parse de variável {value}...")

        variables = value.split("$")[1:]

        if "self" == variables[0]:
            pass  # this method will be implemented on version 1.1.0
        elif "version" == variables[0]:
            pass  # this method will be implemented on version 1.1.0
        else:
            for variable in variables:
                var = variable.replace("/", "")
                v = self._change_value(self._search_variable(var))

                if _is_regex(v):
                    v = v.replace("^", "").replace("$", "")

                value = value.replace(f"${var}", v)

        return value


def test():
    pass



from dataclasses import dataclass

from src.core.actions.executable import Executable
from src.core.models import ActionExecution
from src.core.template_models import Branch, Pattern


@dataclass
class DeleteBranchParameters:
    source: Branch
    pattern: Pattern


class DeleteBranch(Executable):
    parameters: DeleteBranchParameters

    def __init__(self, action_execution: ActionExecution):
        super().__init__(action_execution)
        action_parameters = self.action_execution.parameters

        self.parameters = DeleteBranchParameters(
            source=Branch(action_parameters.get("source", {})),
            pattern=Pattern(action_parameters.get("pattern", {})),
        )

    def execute(self):
        pass  # this method will be implemented on version 1.1.0



"""
git-phoenix setup.py
"""
import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# Package version
VERSION = "1.1.1"

setup(
    name="git-phoenix",
    version=VERSION,
    description="A dynamic git branching tool",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/victoraugustofd/git-phoenix",
    author="victoraugustofd",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "GitPython==3.1.24",
        "regex==2021.4.4",
        "argparse==1.4.0",
        "jsonschema==4.0.1",
        "coloredlogs==15.0.1",
        "questionary==1.10.0",
        "click==8.0.1",
    ],
    python_requires=">=3.7",
    entry_points={
        "console_scripts": [
            "git-phoenix=src.git_phoenix:main",
            "git-px=src.git_phoenix:main",
        ]
    },
)



import inspect
import sys

from src.core.px_git import get_config


def is_mod_function(mod, func):
    return inspect.isfunction(func) and inspect.getmodule(func) == mod


def list_functions(mod):
    return [
        func.__name__
        for func in mod.__dict__.values()
        if is_mod_function(mod, func)
    ]


def raiser(ex: BaseException):
    raise ex


def capitalize_first_letter(word):
    return word[:1].upper() + word[1:]


def get_template():
    return get_config("phoenix.template-path")


def px_print(msg):
    print(msg, file=sys.stderr)
